/* ========================================================================= *
 * stepper.c -- time stepping routines for EZ-pipe (discrete version).   
 *
 * Copyright (C) 2011 Dwight Barkley                 
 *
 * Note: this code is written for speed. There is also a version in which the
 * function f is slightly more readable. Even this version could be speed up
 * slightly more (by dealing more efficiently with ALPHA(u) for example).
 *
 * ========================================================================= */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "pipe.h"
#include "stepper.h"

static double *Sigma_u;
static double *Sigma_q;
static double two_plus_beta;
static double four_plus_beta;
static double one_over_two_minus_gamma;

static void Impose_boundary_conditions (void);

/* ========================================================================= */

double f (double u, double q)
{
  double  return_value;
  double  alpha = ALPHA (u);
  double  Q1 = alpha * one_over_two_minus_gamma;

  if (q < Q1) {
    return_value = gamma * q;
  }
  else if (q < 1.0) {
    return_value = 2.0 * q - alpha;
  }
  else if (q < ((four_plus_beta - alpha - gamma * Q1) / two_plus_beta)) {
    return_value = -two_plus_beta * q + four_plus_beta - alpha;
  }
  else {
    return_value = gamma * Q1;
  }

  return return_value;
}

/* ========================================================================= */

void Step (void)
{
  double  du;
  register int i;

  Impose_boundary_conditions ();

  for (i = 1; i <= nx; i++) {
    Sigma_u[i] = -c_param * (u[i] - u[i - 1]);
    Sigma_q[i] = d_param * (q[i - 1] - 2 * q[i] + q[i + 1]);
  }

  for (i = 1; i <= nx; i++) {
    du = epsilon1 * (1.0 - u[i]) - epsilon2 * q[i] * u[i] + Sigma_u[i];
    q[i] += Sigma_q[i];
    /* 2 iterates of map */
    q[i] = f (u[i], q[i]);
    q[i] = f (u[i], q[i]);
    u[i] += du;
  }

  /* advect forward 1 grid point */
  u[0] = u[nx - 1];
  q[0] = q[nx - 1];
  for (i = nx - 1; i >= 0; i--) {
    u[i + 1] = u[i];
    q[i + 1] = q[i];
  }
}

/* ========================================================================= */

void Step_ini (void)
{
  Sigma_u = (double *) calloc (nx + 2, sizeof (double));
  Sigma_q = (double *) calloc (nx + 2, sizeof (double));

  two_plus_beta = (2. + beta);
  four_plus_beta = (4. + beta);
  one_over_two_minus_gamma = 1. / (2.0 - gamma);

  Impose_boundary_conditions ();
}

/* ========================================================================= */

void Step_fin (void)
{
  free (Sigma_u);
  free (Sigma_q);
}

/* ========================================================================= */

void Impose_boundary_conditions (void)
{
  /* Average periodic directions (mostly this shouldn't be necessary) */
  u[1] = u[nx] = 0.5 * (u[nx] + u[1]);
  q[1] = q[nx] = 0.5 * (q[nx] + q[1]);

  /* Set fictitious points for periodic boundary conditions */
  u[0] = u[nx - 1];
  u[nx + 1] = u[2];

  q[0] = q[nx - 1];
  q[nx + 1] = q[2];
}

/* ========================================================================= */
