/* ========================================================================= *
 *                                                                           *
 *                                 EZ-Pipe                                   *
 *                      Model simulations of pipe flow                       *
 *                       -- The Continuous Model --                          *
 *                               Version 0.3                                 *
 *                      Copyright (C) 2011    Dwight Barkley                 *
 *                                            D.Barkley@warwick.ac.uk        *
 *                                                                           *
 *    EZ-Pipe is free software; you can redistribute it and/or modify        *
 *    it under the terms of the GNU General Public License V3.0 as           *
 *    published by the Free Software Foundation.                             *
 *                                                                           *
 *    EZ-Pipe is distributed in the hope that it will be useful,             *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *    GNU General Public License for more details.                           *
 *                                                                           *
 * ========================================================================= */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include "pipe.h"
#include "stepper.h"
#include "analyzer.h"
#include "plotter.h"

/* Global variables *
 * ---------------- */
double *u, *q;
double  r_param, delta, epsilon1, epsilon2, c_param, grid_h, dt;
int     nx, nsteps, istep;

/* Variables and function prototypes for this file only *
 * ---------------------------- ----------------------- */
static double length, ts;
static int plot_step, anal_step;

static void Initialize (void);
static void Finish_up (void);
static void Write_fc (void);
static void Read_ic (void);
static void Generate_ic (void);

/* macro used to skip to end of input line */
#define NEXT_LINE(fp) while(getc(fp)!='\n');

/* ========================================================================= */

int main (void)
{
  Initialize ();

  for (istep = 0; istep < nsteps; istep++) {
    if (Event_check ()) break;
    if (plot_step && (istep % plot_step) == 0) Draw ();
    if (anal_step && (istep % anal_step) == 0) Analyze (istep);
    Step ();
  }

  Finish_up ();
  return 0;
}

/* ========================================================================= */

void Initialize (void)
{
  double  p_in;
  FILE   *fp;

  /* ------------ Read parameters from task.dat ------------ */

  if ((fp = fopen ("task.dat", "r")) == NULL) {
    fprintf (stderr, "Cannot open task file: task.dat \n");
    exit (1);
  }
  else {
    fscanf (fp, "%lg", &p_in);    NEXT_LINE (fp);    r_param = p_in;
    fscanf (fp, "%lg", &p_in);    NEXT_LINE (fp);    delta = p_in;
    fscanf (fp, "%lg", &p_in);    NEXT_LINE (fp);    epsilon1 = p_in;
    fscanf (fp, "%lg", &p_in);    NEXT_LINE (fp);    epsilon2 = p_in;
    NEXT_LINE (fp);
    fscanf (fp, "%d", &nx);    NEXT_LINE (fp);
    fscanf (fp, "%lg", &p_in);    NEXT_LINE (fp);    length = p_in;
    fscanf (fp, "%lg", &p_in);    NEXT_LINE (fp);    c_param = p_in;
    fscanf (fp, "%lg", &p_in);    NEXT_LINE (fp);    ts = p_in;
    NEXT_LINE (fp);
    fscanf (fp, "%d", &nsteps);    NEXT_LINE (fp);
    fscanf (fp, "%d", &plot_step);    NEXT_LINE (fp);
    fscanf (fp, "%d", &anal_step);    NEXT_LINE (fp);
    fclose (fp);
  }

  grid_h = length / (nx - 1);
  dt = ts * (grid_h * grid_h / 2.);

  printf ("\n\nParameters: \n");
  printf ("r        = %g\n", r_param);
  printf ("delta    = %g\n", delta);
  printf ("epsilon1 = %g\n", epsilon1);
  printf ("epsilon2 = %g\n", epsilon2);
  printf ("c_param  = %g\n", c_param);
  printf ("length   = %g\n", length);

  printf ("\nNumber of grid points   = %d\n", nx);
  printf ("Number of time steps    = %d\n", nsteps);
  printf ("With: dt   = %-10g grid_h = %-10g\n", dt, grid_h);
  printf ("time steps per plot     = %d\n", plot_step);
  printf ("time steps per analysis = %d\n", anal_step);

  /* ------------ Final initialization ------------ */

  u = (double *) calloc (nx + 2, sizeof (double));
  q = (double *) calloc (nx + 2, sizeof (double));

  Read_ic ();
  Step_ini ();
  Analyze_ini ();
  Draw_ini ();
}

/* ========================================================================= */

void Finish_up (void)
{
  QuitX ();
  Step_fin ();
  Analyze_fin ();
  Write_fc ();
}

/* ========================================================================= */

void Write_fc (void)
{
  /* Write final condition file */

  FILE   *fp;
  time_t  tt1;

  time (&tt1);
  fp = fopen ("fc.dat", "w");

  fprintf (fp, "Parameters: r, delta, epsilon1, epsilon2: ");
  fprintf (fp, "%g, %g, %g, %g\n", r_param, delta, epsilon1, epsilon2);
  fprintf (fp, "nx, L, c_param = ");
  fprintf (fp, "%d, %g, %g \n", nx, length, c_param);
  fprintf (fp, "File written: %s", ctime (&tt1));
  fprintf (fp, "Comments: \n");
  fprintf (fp, "\n");
  fprintf (fp, "\n");
  fprintf (fp, "\n");
  fprintf (fp, "\n");
  fprintf (fp, "\n");

  fprintf (fp, "Ascii values of u and q follow\n");
  for (int i = 1; i <= nx; i++) {
    fprintf (fp, "%g  %g\n", u[i], q[i]);
  }

  fclose (fp);
}

/* ========================================================================= */

void Read_ic (void)
{
  /* Reads initial condition file (ic.dat) if it exists, otherwise calls
     Generate_ic() to generate new initial condition. */

  double *u_ic, *q_ic, rx;
  int     nx_ic, i_ic, i;
  char    dummy;
  FILE   *fp;

  if ((fp = fopen ("ic.dat", "r")) == NULL) {
    Generate_ic ();
    return;
  }

  /* Read nx_ic etc following = sign on second line of file */
  NEXT_LINE (fp);
  while ((dummy = getc (fp)) != '=');
  fscanf (fp, "%d", &nx_ic);

  /* Skip to 11th line */
  for (i = 0; i < 9; i++)
    NEXT_LINE (fp);

  printf ("\nReading ic.dat with nx = %d... \n\n", nx_ic);

  /* Allocate temporary memory and read from file */
  q_ic = (double *) malloc ((unsigned) (nx_ic) * sizeof (double));
  u_ic = (double *) malloc ((unsigned) (nx_ic) * sizeof (double));

  for (i = 0; i < nx_ic; i++) {
    fscanf (fp, "%lg  %lg", &u_ic[i], &q_ic[i]);
    fscanf (fp, "\n");
  }

  rx = (nx_ic - 1.0) / (nx - 1.0);
  for (i = 1; i <= nx; i++) {
    i_ic = (int) (rx * (i - 1));
    u[i] = u_ic[i_ic];
    q[i] = q_ic[i_ic];
  }

  free (q_ic);
  free (u_ic);
}

/* ========================================================================= */

void Generate_ic (void)
{
  /* Generate new initial condition */

  printf ("\nGenerating initial condition \n\n");

  for (int i = 1; i <= nx; i++) {
    u[i] = 1;
    q[i] = 0.;
  }
  Generate_puff ();
}

/* ========================================================================= */

void Generate_puff (void)
{
  /* Generate puff near left side of domain */

  int     start = (nx - 1) / 10;
  int     stop = min (start + 20, nx);

  for (int i = start; i <= stop; i++) {
    q[i] += 0.5 + 0.05 * rand () / RAND_MAX;
  }
}

/* ========================================================================= */

void Generate_hole (void)
{
  /* Generate hole near left side of domain */

  int     start = (nx - 1) / 10;
  int     stop = min (start + 20, nx);

  for (int i = start; i <= stop; i++) {
    q[i] = 0.;
  }
}

/* ========================================================================= */
