/* ========================================================================= *
 * plotter.c -- X11/OpenGL graphics routines for EZ-pipe.  
 *
 * Copyright (C) 2011 Dwight Barkley                 
 * ========================================================================= */

#include "pipe.h"
#include "plotter.h"

#if !GRAPHICS			/* dummy functions */

void Draw (void)
{
}

void Draw_ini (void)
{
}

int Event_check (void)
{
  return 0;
}

void QuitX (void)
{
}

#else /* actual graphics */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>
#include <GL/gl.h>
#include <GL/glx.h>

#define WINDOW_TITLE "Pipe"

#define WINSIZE      300	/* Window is square of this size in pixels. */
#define PLOT_SIZE    1.1	/* This controls the size of the simulation
				   volume in the view port: >1.0 for larger
				   size, <1.0 for smaller size. */
#define WINX         0		/* Window location, in pixels, from screen left 
				 */
#define WINY         0		/* Window location, in pixels, from screen top. 
				 */
#define WM_CTRLS_POS 0		/* 1 for WM_CTRL_POS to control position */

#define BACKGROUND   1.0	/* Background color (R=G=B=BACKGROUND, so 0.0
				   gives BLACK, 1.0 gives WHITE) */

#define START_PAUSED 1		/* If 1 then window is opened in paused mode
				   showing initial condition. */

/* leave the following alone */

#define TRUE             1
#define FALSE            0

#define SPACE            0	/* These are used to determine which field */
#define PP               1	/* (if any) is being plotted */
#define NO_FIELD        -1

#define MODE_SIMULATING  1
#define MODE_VIEWING     2


/* 
 * Global variables for this file only 
 * ----------------------------------- */
static Display *theDisplay;
static Window theWindow;

static int field;		/* Field being viewed */
static int ezmode;		/* Mode flag */

static double half_width, half_height;

static double plot_length[2];	/* Lengths of the simulation volume in graphics 
				   coordinates (see Draw_ini()). */
/* 
 * Private functions 
 * ----------------- */
static void Restart (void);
static void Pause (void);
static void View_space (void);
static void View_pp (void);
static void View_no_field (void);
static void myReshape (int w, int h);
static void Mover (int m_x);
static void InitX (int winx, int winy, int width, int height);
static Bool WaitForNotify (Display * d, XEvent * e, char *arg);
static void Save_image (void);

/* ========================================================================= */

void Draw (void)
{
  double  rect_h = plot_length[0] / (nx - 1);
  int     i;
  /* Main plotting routine */

  glClear (GL_COLOR_BUFFER_BIT);

#define PX(x) ((rect_h*((x)-1.))-half_width)

#define PU(y) ( (0.9*(y) - 1.)*half_height )
#define PQ(y) ( (0.5*(y))*half_height )

#define PPU(y) ( 0.5*(y) - 0.5 * half_width)
#define PPQ(y) ( 0.5*(y) - 0.75 * half_height)

  if (field == SPACE) {

    glLineWidth (2);
    glBegin (GL_LINE_STRIP);
    glColor3f (0., 0., 1.);
    for (i = 0; i <= nx; i++) {
      glVertex2f (PX (i), PU (u[i]));
    }
    glEnd ();

    glLineWidth (2);
    glBegin (GL_LINE_STRIP);
    glColor3f (1., 0., 0.);
    for (i = 0; i <= nx; i++) {
      glVertex2f (PX (i), PQ (q[i]));
    }
    glEnd ();

  }

  if (field == PP) {		/* Phase portrait */

    glBegin (GL_LINES);
    glColor3f (0, 0, 0);
    glVertex2f (PPU (-10.), PPQ (0.));
    glVertex2f (PPU (10.), PPQ (0.));
    glEnd ();
    glBegin (GL_LINES);
    glColor3f (0, 0, 0);
    glVertex2f (PPU (0.), PPQ (-10.));
    glVertex2f (PPU (0.), PPQ (10.));
    glEnd ();

    glBegin (GL_LINE_STRIP);
    glLineWidth (3);
    glColor3f (1., 0., 0.);
    for (i = 0; i <= nx; i++) {
      glVertex2f (PPU (u[i]), PPQ (q[i]));
    }
    glEnd ();

  }

  glXSwapBuffers (theDisplay, theWindow);
}

/* ========================================================================= */

static void Restart (void)
{
  /* Return to MODE_SIMULATING. This has the effect of restarting the
     simulation via a return from Event_check(). */

  if (ezmode == MODE_VIEWING) {
    ezmode = MODE_SIMULATING;
  }
}

/* ========================================================================= */

static void Pause (void)
{
  if (ezmode == MODE_SIMULATING) {
    ezmode = MODE_VIEWING;
  }
}

/* ========================================================================= */

static void View_space (void)
{
  field = SPACE;
  if (ezmode != MODE_SIMULATING) {
    Draw ();
  }
}

/* ========================================================================= */

static void View_pp (void)
{
  field = PP;
  if (ezmode != MODE_SIMULATING) {
    Draw ();
  }
}

/* ========================================================================= */

static void View_no_field (void)
{
  field = NO_FIELD;
  if (ezmode != MODE_SIMULATING) {
    Draw ();
  }
}

/* ========================================================================= */

static void myReshape (int w, int h)
{
  /* half_width and half_height define the area viewed in GL.  In general if
     these are large, then simulation area will appear small, and vice versa.
     PLOT_SIZE allows adjustment of this without changing any of the code
     below. */

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glOrtho (-PLOT_SIZE * half_width, PLOT_SIZE * half_width,
	   -PLOT_SIZE * half_height, PLOT_SIZE * half_height, -20., 20.);
  glMatrixMode (GL_MODELVIEW);
  glViewport (0, 0, w, h);
}

/* ========================================================================= */

void Draw_ini (void)
{
  /* Initialize everything necessary for plotting.  */

  plot_length[0] = 2.0;
  plot_length[1] = 1.0;

  half_width = 0.5 * plot_length[0];
  half_height = 0.5 * plot_length[1];

  /* Create X window and prepares for use by OpenGL */
  InitX (WINX, WINY, WINSIZE * plot_length[0], WINSIZE * plot_length[1]);

  /* Set the background color.  Here: red=green=blue=BACKGROUND.  One can set
     any values one chooses so long as 0 <= red, green, blue <=1 */
  glClearColor (BACKGROUND, BACKGROUND, BACKGROUND, 0.0);

  /* Set starting mode */
  if (START_PAUSED) {
    ezmode = MODE_VIEWING;
  }
  else {
    ezmode = MODE_SIMULATING;
  }

}

/* ========================================================================= */

void Mover (int m_x)
{
  /* Moves solution in response to a key press */
  if (m_x > 0) {
    for (int i = nx - 1; i > 0; i--) {
      u[i + 1] = u[i];
      q[i + 1] = q[i];
    }
    u[1] = u[nx];
    q[1] = q[nx];
  }
  else {
    for (int i = 1; i < nx; i++) {
      u[i] = u[i + 1];
      q[i] = q[i + 1];
    }
    u[nx] = u[1];
    q[nx] = q[1];
  }

  if (ezmode != MODE_SIMULATING) {
    Draw ();
  }
}

/* ========================================================================= */

static void InitX (int winx, int winy, int width, int height)
{
  /* Initializes X and opens a window. */

  static XVisualInfo *theVisualInfo;
  static GLXContext theGLXContext;
  static Colormap theColormap;
  static int theScreen;
  static int theDepth;
  static int theDWidth;
  static int theDHeight;
  static char *theDisplayName = NULL;
  static XEvent event;
  static Atom del_atom;
  static XSizeHints theSizeHints;
  static XSetWindowAttributes theSWA;
  static char *name = WINDOW_TITLE;
  static XTextProperty theWindowName, theIconName;
  static int num1, num2;
  static int list[] = { GLX_RGBA,
    GLX_DOUBLEBUFFER,
    GLX_RED_SIZE, 1,
    GLX_GREEN_SIZE, 1,
    GLX_BLUE_SIZE, 1,
    GLX_DEPTH_SIZE, 1,
    None
  };

  /* DOUBLE is the important one.  Perhaps need to add error checking when
     DOUBLE_BUFFER not available.  In the aux library the first entry in list
     was GLX_LEVEL, 0, */

  /* Open the display */
  if ((theDisplay = XOpenDisplay (NULL)) == NULL) {
    fprintf (stderr,
	     "ERROR: Could not open a connection to X on display %s\n",
	     XDisplayName (theDisplayName));
    exit (1);
  }
  if (!glXQueryExtension (theDisplay, &num1, &num2)) {
    fprintf (stderr,
	     "ERROR: No glx extension on display %s\n",
	     XDisplayName (theDisplayName));
    exit (1);
  }

  theScreen = DefaultScreen (theDisplay);
  theDepth = DefaultDepth (theDisplay, theScreen);
  theDWidth = DisplayWidth (theDisplay, theScreen);
  theDHeight = DisplayHeight (theDisplay, theScreen);

  if (!(theVisualInfo = glXChooseVisual (theDisplay, theScreen, list))) {
    fprintf (stderr, "ERROR: Couldn't find visual");
    exit (-1);
  }
  if (!(theGLXContext = glXCreateContext (theDisplay, theVisualInfo,
					  None, GL_TRUE))) {
    /* Last parameter indicates that, if possible, then render directly to
       graphics hardware and bypass the X server. This should be faster. */
    fprintf (stderr, "ERROR: Can not create a context!\n");
    exit (-1);
  }

  theColormap = XCreateColormap (theDisplay,
				 RootWindow (theDisplay,
					     theVisualInfo->screen),
				 theVisualInfo->visual, AllocNone);
  /* AllocAll would generate a BadMatch.  */

  if (!(theColormap)) {
    fprintf (stderr, "ERROR: couldn't create Colormap\n");
    exit (-1);
  }
  theSWA.colormap = theColormap;
  theSWA.border_pixel = 0;
  theSWA.event_mask = (EnterWindowMask | KeyPressMask | StructureNotifyMask |
		       ButtonPressMask | ButtonReleaseMask | ExposureMask |
		       PointerMotionMask);

  theWindow = XCreateWindow (theDisplay,
			     RootWindow (theDisplay, theVisualInfo->screen),
			     winx, winy, width, height, 0,
			     theVisualInfo->depth, InputOutput,
			     theVisualInfo->visual,
			     CWBorderPixel | CWColormap | CWEventMask,
			     &theSWA);

  if (!(theWindow)) {
    fprintf (stderr, "ERROR: couldn't create X window\n");
    exit (-1);
  }

  /* Set standard window properties.  */

  XStringListToTextProperty (&name, 1, &theWindowName);
  XStringListToTextProperty (&name, 1, &theIconName);

  theSizeHints.base_width = width;
  theSizeHints.base_height = height;
  theSizeHints.min_aspect.x = width;	/* Maintain x:y ratio */
  theSizeHints.max_aspect.x = width;
  theSizeHints.min_aspect.y = height;
  theSizeHints.max_aspect.y = height;

  theSizeHints.flags = PSize | PAspect;

  if (!(WM_CTRLS_POS))
    theSizeHints.flags |= USPosition;
  /* Setting USPosition here seems to make the WM honor the x and y specified
     by XCreateWindow above.  Not setting this should give control of position
     to the WM.  Note that the root window of an application is special in that 
     the WM has special privileges over its properties so this may not work on
     all platforms.  */

  XSetWMProperties (theDisplay, theWindow, &theWindowName, &theIconName,
		    NULL, 0, &theSizeHints, NULL, NULL);

  /* Express interest in WM killing this application */
  if ((del_atom = XInternAtom (theDisplay, "WM_DELETE_WINDOW", TRUE)) != None) {
    XSetWMProtocols (theDisplay, theWindow, &del_atom, 1);
  }

  XMapWindow (theDisplay, theWindow);
  XIfEvent (theDisplay, &event, WaitForNotify, (char *) theWindow);

  glXMakeCurrent (theDisplay, theWindow, theGLXContext);

  /* Print useful information. I suggest printing this at this at least once */
  if (0) {
    printf ("%s version %d of the X Window System, X%d R%d\n",
	    ServerVendor (theDisplay),
	    VendorRelease (theDisplay),
	    ProtocolVersion (theDisplay), ProtocolRevision (theDisplay));

    if (theDepth == 1) {
      printf ("Color plane depth...........%d (monochrome)\n", theDepth);
    }
    else {
      printf ("Color plane depth...........%d \n", theDepth);
    }

    printf ("Display Width...............%d \n", theDWidth);
    printf ("Display Height..............%d \n", theDHeight);
    printf ("The display %s\n", XDisplayName (theDisplayName));
  }
}

/* ========================================================================= */

static Bool WaitForNotify (Display * d, XEvent * e, char *arg)
{
  /* As seen in the Porting Guide. */
  return (e->type == MapNotify) && (e->xmap.window == (Window) arg);
}

/* ========================================================================= */

void QuitX (void)
{
  XCloseDisplay (theDisplay);
  return;
}

/* ========================================================================= */

static void Save_image (void)
{
  static int num = 1000;
  char    cmd[40];

  printf ("taking snapshot %d\n", num);

  glFlush ();
  sprintf (cmd, "import -window Pipe snap.%d.png", num++);
  system (cmd);
}

/* ========================================================================= */

int Event_check (void)
{
  static XEvent theEvent;
  static KeySym theKeySym;
  static int theKeyBufferMaxLen = 64;
  static char theKeyBuffer[65];
  static XComposeStatus theComposeStatus;

  /* X Event Loop If there is something in the queue then each event is
     processed in turn. When the queue is empty, and the mode (which may have
     been changed by the events just processed) is MODE_SIMULATING then control 
     is returned (presumably to the main loop in main()).  However, when the
     queue is empty and the mode is either MODE_ROTATING or MODE_VIEWING then
     XNextEvent is called which blocks until the next X event is received (eg.
     the space bar being pressed which sets the mode back to MODE_SIMULATING
     and so control returns to main()). */

  while (XPending (theDisplay) || (ezmode != MODE_SIMULATING)) {

    XNextEvent (theDisplay, &theEvent);

    switch (theEvent.type) {	/* switch according to X event */

    case KeyPress:		/* A KeyPress event has occurred. */

      XLookupString ((XKeyEvent *) & theEvent, theKeyBuffer,
		     theKeyBufferMaxLen, &theKeySym, &theComposeStatus);

      switch (theKeySym) {	/* switch according to the pressed key */

      case XK_Escape:
	exit (0);		/* hard exit, nothing saved */

      case XK_Q:
      case XK_q:
	return (1);

      case XK_P:
      case XK_p:
	Pause ();
	Draw ();
	break;

      case XK_space:
	Restart ();
	break;

      case XK_U:
      case XK_u:
	View_space ();
	break;

      case XK_V:
      case XK_v:
	View_pp ();
	break;

      case XK_N:
      case XK_n:
	View_no_field ();
	break;

      case XK_R:
	r_param += 0.01;
	printf ("r = %g\n", r_param);
	break;
      case XK_r:
	r_param -= 0.01;
	printf ("r = %g\n", r_param);
	break;

      case XK_C:
	c_param += 0.01;
	printf ("c = %g\n", c_param);
	break;
      case XK_c:
	c_param -= 0.01;
	printf ("c = %g\n", c_param);
	break;

      case XK_E:
	epsilon1 += 0.5;
	printf ("epsilon1=%g\n", epsilon1);
	break;
      case XK_e:
	epsilon2 -= 0.5;
	printf ("epsilon1=%g\n", epsilon1);
	break;

      case XK_x:
	Generate_puff ();
	break;
      case XK_X:
	Generate_hole ();
	break;

      case XK_l:
	for (int i = 1; i <= nx; i++) {
	  u[i] = 1;
	  q[i] = 0.;
	}
	break;

      case XK_S:
      case XK_s:
	Save_image ();
	break;

      case XK_Right:
	Mover (1);
	break;			/* move +x */
      case XK_Left:
	Mover (-1);
	break;			/* move -x */

      }				/* switch(theKeySym) */
      break;

    case EnterNotify:
      /* This case is necessary for window managers which do not set keyboard
         focus to theWindow when the pointer enters theWindow. */
      XSetInputFocus (theDisplay, theWindow, RevertToPointerRoot,
		      CurrentTime);
      break;

    case Expose:
      /* Window mapped for the first time and if you uncover some part of the
         window. If you start paused and you see nothing in the window then its 
         possible that the problem is that the first Expose event is not being
         caught for some reason. */
      Draw ();
      break;

    case ConfigureNotify:
      /* Window size is changed by the user (or the window is initially
         opened). Note that InitX contains code that constrains the window to
         be square. */
      myReshape (theEvent.xconfigure.width, theEvent.xconfigure.height);
      break;

    }				/* switch (theEvent.type) */

  }				/* while (XPending(theDisplay) || (ezmode !=
				   MODE_SIMULATING)) */

  return (0);
}

/* ========================================================================= */

#endif /* end actual graphics */
